/*
  File di esempio con programma parallelo permoltiplicare due matrici.

  DISCLAIMER: il programma e` stato fatto in un giorno e le
  ottimizzazioni in un altro, quindi con alta probabilita` ci sara`
  qualche errore. Guardate soprattutto l'allocazione dinamica della
  memoria per le matrici, l'uso di hpm e le tecniche di
  ottimizzazione.

  Buon lavoro...
*/

#include <stdio.h>
#include <mpi.h>
#include <stdlib.h>
#include "libhpm.h"

const int N=512;

/* allocate memory for an RxN  matrix */
float** createM(int R, int C){
  int n=R*C; int i=0;
  float* data=calloc(n, sizeof(float));
  float** M=malloc(R * sizeof(float*));
  
  for (; i<R; ++i)
    M[i]=data+i*C;

  return M;
}

/* read square matrix from file */
float** readM(char* filename, int L){
  float** M=createM(L,L);
  FILE* fin=fopen(filename, "r");

  fread(M[0], sizeof(float), L*L, fin);
  fclose(fin);

  return M;
}

/* free memory allocated for a matrix */
void freeM(float **M){
  free(M[0]);
  free(M);
}

/* transpose of a matrix */
float** transM(float **A, int R, int C){
  float** At=createM(C,R);
  int j, i=0;

  for(; i<R; ++i)
    for(j=0; j<C; ++j)
      At[j][i]=A[i][j];

  return At;
}


/* product */
void Prod(float **c, float **r, float **P, int l){
  int i, j, k;
  for(i=0; i<N; ++i)
    for(j=0; j<N; ++j)
      for(k=0; k<l; ++k)
	P[i][j]+=c[k][i]*r[k][j];
}

/* product */
void Prod2(float **c, float **r, float **P, int l){
  int i, j, k;
  for(k=0; k<l; ++k)
    for(i=0; i<N; ++i)
      for(j=0; j<N; ++j)
	P[i][j]+=c[k][i]*r[k][j];
}

/* product */
void Prod3a(float **c, float **r, float **P, int l){
  int i, j, k;
  for(k=0; k<l; ++k)
    for(i=0; i<N; ++i)
      for(j=0; j<(N>>3); ++j){
	int jj=j<<3;
	P[i][jj+0]+=c[k][i]*r[k][j+0];
	P[i][jj+1]+=c[k][i]*r[k][j+1];
	P[i][jj+2]+=c[k][i]*r[k][j+2];
	P[i][jj+3]+=c[k][i]*r[k][j+3];
	P[i][jj+4]+=c[k][i]*r[k][j+4];
	P[i][jj+5]+=c[k][i]*r[k][j+5];
	P[i][jj+6]+=c[k][i]*r[k][j+6];
	P[i][jj+7]+=c[k][i]*r[k][j+7];
      }
}


/* product */
void Prod3b(float **c, float **r, float **P, int l){
  int i, j, k;
  for(k=0; k<l; ++k)
    for(i=0; i<N; ++i)
      for(j=0; j<(N>>3); ++j){
	int jj=j<<3;
	float *PP=&(P[i][jj]);
	float cc=c[k][i];
	float *rr=&(r[k][jj]);
	PP[0]+=cc*(rr[0]);
	PP[1]+=cc*(rr[1]);
	PP[2]+=cc*(rr[2]);
	PP[3]+=cc*(rr[3]);
	PP[4]+=cc*(rr[4]);
	PP[5]+=cc*(rr[5]);
	PP[6]+=cc*(rr[6]);
	PP[7]+=cc*(rr[7]);
      }
}




/* product */
void Prod4(float **c, float **r, float **P, int l){
  int i, j, k;
  for(k=0; k<l; ++k)
    for(i=0; i<N; ++i)
      for(j=0; j<(N>>4); ++j){
	int jj=j<<4;
	float *PP=&(P[i][jj]);
	float cc=c[k][i];
	float *rr=&(r[k][jj]);
	PP[0]+=cc*(rr[0]);
	PP[1]+=cc*(rr[1]);
	PP[2]+=cc*(rr[2]);
	PP[3]+=cc*(rr[3]);
	PP[4]+=cc*(rr[4]);
	PP[5]+=cc*(rr[5]);
	PP[6]+=cc*(rr[6]);
	PP[7]+=cc*(rr[7]);
	PP[8]+=cc*(rr[8]);
	PP[9]+=cc*(rr[9]);
	PP[10]+=cc*(rr[10]);
	PP[11]+=cc*(rr[11]);
	PP[12]+=cc*(rr[12]);
	PP[13]+=cc*(rr[13]);
	PP[14]+=cc*(rr[14]);
	PP[15]+=cc*(rr[15]);
      }
}

/* product */
void Prod5(float **c, float **r, float **P, int l){
  int i, j, k;
  for(k=l-1; k>=0; --k)
    for(i=N-1; i>=0; --i)
      for(j=(N>>5)-1; j>=0; --j){
	int jj=(j<<5);
	float *PP=&(P[i][jj]);
	float cc=c[k][i];
	float *rr=&(r[k][jj]);
	PP[0]+=cc*(rr[0]);
	PP[1]+=cc*(rr[1]);
	PP[2]+=cc*(rr[2]);
	PP[3]+=cc*(rr[3]);
	PP[4]+=cc*(rr[4]);
	PP[5]+=cc*(rr[5]);
	PP[6]+=cc*(rr[6]);
	PP[7]+=cc*(rr[7]);
	PP[8]+=cc*(rr[8]);
	PP[9]+=cc*(rr[9]);
	PP[10]+=cc*(rr[10]);
	PP[11]+=cc*(rr[11]);
	PP[12]+=cc*(rr[12]);
	PP[13]+=cc*(rr[13]);
	PP[14]+=cc*(rr[14]);
	PP[15]+=cc*(rr[15]);
      }
}

/* product */
void Prod6(float **c, float **r, float **P, int l){
  int i, j, k;
  for(k=l-1; k>=0; --k)
    for(i=N-1; i>=0; --i)
      for(j=(N>>5)-1; j>=0; --j){
	int jj=(j<<5);
	float *PP=&(P[i][jj]);
	float cc=c[k][i];
	float *rr=&(r[k][jj]);
	PP[0]+=cc*(rr[0]);
	PP[1]+=cc*(rr[1]);
	PP[2]+=cc*(rr[2]);
	PP[3]+=cc*(rr[3]);
	PP[4]+=cc*(rr[4]);
	PP[5]+=cc*(rr[5]);
	PP[6]+=cc*(rr[6]);
	PP[7]+=cc*(rr[7]);
	PP[8]+=cc*(rr[8]);
	PP[9]+=cc*(rr[9]);
	PP[10]+=cc*(rr[10]);
	PP[11]+=cc*(rr[11]);
	PP[12]+=cc*(rr[12]);
	PP[13]+=cc*(rr[13]);
	PP[14]+=cc*(rr[14]);
	PP[15]+=cc*(rr[15]);
	PP[16]+=cc*(rr[16]);
	PP[17]+=cc*(rr[17]);
	PP[18]+=cc*(rr[18]);
	PP[19]+=cc*(rr[19]);
	PP[20]+=cc*(rr[20]);
	PP[21]+=cc*(rr[21]);
	PP[22]+=cc*(rr[22]);
	PP[23]+=cc*(rr[23]);
	PP[24]+=cc*(rr[24]);
	PP[25]+=cc*(rr[25]);
	PP[26]+=cc*(rr[26]);
	PP[27]+=cc*(rr[27]);
	PP[28]+=cc*(rr[28]);
	PP[29]+=cc*(rr[29]);
	PP[30]+=cc*(rr[30]);
	PP[31]+=cc*(rr[31]);
      }
}


/* product */
void ProdV2(float **ct, float **rt, float **P, int l){
  int i, j, k, jj;
  float *PP, PP0, PP1, PP2, PP3, PP4, PP5, PP6, PP7;
  float cc;

  for(i=N-1; i>=0; --i)
    for(j=(N>>3)-1; j>=0; --j){
      jj=j<<3;
      PP=&(P[i][jj]);
      PP0=0; PP1=0; PP2=0; PP3=0; PP4=0; PP5=0; PP6=0; PP7=0;
      for(k=l-1; k>=0; --k){

	cc=ct[i][k];

	PP0+=cc*rt[jj][k];
	PP1+=cc*rt[jj+1][k];
	PP2+=cc*rt[jj+2][k];
	PP3+=cc*rt[jj+3][k];
	PP4+=cc*rt[jj+4][k];
	PP5+=cc*rt[jj+5][k];
	PP6+=cc*rt[jj+6][k];
	PP7+=cc*rt[jj+7][k];

      }
      PP[0]=PP0;
      PP[1]=PP1;
      PP[2]=PP2;
      PP[3]=PP3;
      PP[4]=PP4;
      PP[5]=PP5;
      PP[6]=PP6;
      PP[7]=PP7;
    }
}


int main (int argc, char *argv[]){
  float **A, **B, **C, **At;
  int rank, size;

  MPI_Init(&argc, &argv);
  MPI_Comm_rank(MPI_COMM_WORLD, &rank);
  MPI_Comm_size(MPI_COMM_WORLD, &size);
  hpmInit(rank, "MatMul" );

  /* only the master -- read data from disk */
  if(rank==0){
    hpmStart(1, "Reading from disk and transposing");
    A=readM("mat-a.bin", N);
    B=readM("mat-b.bin", N);
    At=transM(A,N,N);
    C=createM(N,N);
    hpmStop(1);
  }

  /* each processor -- receive data, compute and send back*/
  {
    float **c, **r, **P, **ct, **rt;
    float *pa=0, *pb=0, *pc=0;
    int l=N/size;
    c=createM(l,N);
    r=createM(l,N);
    P=createM(N,N);

    if (rank==0){
      pa=At[0];
      pb=B[0];
      pc=C[0];
    }
    hpmStart(2, "Scattering data");
    MPI_Scatter(pa, N*l, MPI_FLOAT, c[0], N*l, MPI_FLOAT, 0, MPI_COMM_WORLD);
    MPI_Scatter(pb, N*l, MPI_FLOAT, r[0], N*l, MPI_FLOAT, 0, MPI_COMM_WORLD);
    hpmStop(2);
    ct=transM(c,l,N);
    rt=transM(r,l,N);
    hpmStart(3, "Computing");
    ProdV2(ct,rt,P,l);
    hpmStop(3);
    hpmStart(4, "Reducing");
    MPI_Reduce(P[0], pc, N*N, MPI_FLOAT, MPI_SUM, 0, MPI_COMM_WORLD);
    hpmStop(4);

    freeM(c); freeM(r); freeM(P);
    fflush(stdout);
  }


  
  /* only the master -- free the memory and write to disk*/
  if(rank==0){
    FILE* fout=fopen("mat-c.bin", "w");
    hpmStart(5, "Saving to disk");
    fwrite(C[0], sizeof(float), N*N, fout);
    fclose(fout);
    hpmStop(5);

    freeM(A); freeM(B); freeM(C);
  }

  hpmTerminate(rank);

  MPI_Finalize();
  
  return 0;
}
